/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.util;

import fr.amap.commons.util.CallableTask;
import fr.amap.commons.util.Cancellable;
import java.io.File;

/**
 * AMAPVox task based on CallableTask and Cancellable. Returns an array of
 * files.
 *
 * @author Philippe Verley (philippe.verley@ird.fr)
 */
abstract public class AVoxTask extends CallableTask<File[]> implements Cancellable {

    protected final File file;
    protected final int ncpu;
    private boolean cancelled;

    /**
     * Creates a new task with given configuration file and number of threads
     * allocated to the task.
     * @param file, the XML configuration file.
     * @param ncpu, the number of threads allocated to the task.
     */
    public AVoxTask(File file, int ncpu) {
        this.file = file;
        this.ncpu = ncpu;
        cancelled = false;
    }

    /**
     * Initialises the task. Must be light-weighted step.
     *
     * @throws Exception
     */
    abstract public void init() throws Exception;

    /**
     * Return the name of the task.
     *
     * @return the name of the task.
     */
    abstract public String getName();

    /**
     * Return the XML configuration file associated to the task.
     *
     * @return the configuration file.
     */
    public File getFile() {
        return file;
    }

    @Override
    public boolean isCancelled() {
        return cancelled;
    }

    @Override
    public void setCancelled(boolean cancelled) {
        this.cancelled = cancelled;
    }
}
