/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox.voxelisation.output;

import fr.amap.lidar.amapvox.commons.Voxel;
import fr.amap.lidar.amapvox.voxelisation.Voxelization;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import ucar.ma2.Array;
import ucar.ma2.ArrayFloat;
import ucar.ma2.ArrayInt;
import ucar.ma2.DataType;
import ucar.ma2.Index;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFileWriter;
import ucar.nc2.Variable;

/**
 *
 * @author pverley
 */
public class NetCDFVoxelOutput extends AbstractOutput {

    public NetCDFVoxelOutput(Voxelization voxelisation) {
        super(voxelisation);
    }

    @Override
    public File[] write() throws IOException {

        String fileName = voxCfg.getOutputFile().getAbsolutePath();
        if (fileName.endsWith(".vox")) {
            fileName = fileName.substring(0, fileName.length() - 3) + "nc";
        }
        File outputFile = new File(fileName);
        NetcdfFileWriter writer = NetcdfFileWriter.createNew(NetcdfFileWriter.Version.netcdf3, outputFile.getAbsolutePath(), null);

        // add dimensions
        int nx = voxParam.infos.getSplit().x;
        int ny = voxParam.infos.getSplit().y;
        int nz = voxParam.infos.getSplit().z;
        Dimension xDim = writer.addDimension(null, "x", nx);
        Dimension yDim = writer.addDimension(null, "y", ny);
        Dimension zDim = writer.addDimension(null, "z", nz);

        List<Dimension> dims = new ArrayList();
        dims.add(xDim);
        dims.add(yDim);
        dims.add(zDim);

        // add variables
        for (Voxel.OutputVariable variable : Voxel.OutputVariable.values()) {
            if ((!variable.isCoordinateVariable()) && voxParam.isOutputVariableEnabled(variable)) {
                Variable ncvar = writer.addVariable(null, variable.getShortName(), variable.getType(), dims);
                ncvar.addAttribute(new Attribute("units", variable.getUnits()));
                ncvar.addAttribute(new Attribute("long_name", variable.getLongName()));
                ncvar.addAttribute(new Attribute("missing_value", Float.NaN));
            }
        }

        // add global attributes (same as vox file)
        writer.addGroupAttribute(null, new Attribute("min_corner", voxParam.infos.getMinCorner().toString()));
        writer.addGroupAttribute(null, new Attribute("max_corner", voxParam.infos.getMaxCorner().toString()));
        writer.addGroupAttribute(null, new Attribute("split", voxParam.infos.getSplit().toString()));
        writer.addGroupAttribute(null, new Attribute("type", voxParam.infos.getType().toString()));
        writer.addGroupAttribute(null, new Attribute("res", voxParam.infos.getResolution()));
        writer.addGroupAttribute(null, new Attribute("MAX_PAD", voxParam.infos.getMaxPAD()));
        writer.addGroupAttribute(null, new Attribute("LAD_TYPE", voxParam.infos.getLadType().toString()));
        writer.addGroupAttribute(null, new Attribute("build-version", voxParam.infos.getVersion()));

        // create the NetCDF file
        writer.create();

        // write data
        for (Voxel.OutputVariable variable : Voxel.OutputVariable.values()) {
            if ((!variable.isCoordinateVariable()) && voxParam.isOutputVariableEnabled(variable)) {
                Variable ncvar = writer.findVariable(variable.getShortName());
                Array array = (DataType.INT == variable.getType())
                        ? new ArrayInt.D3(nx, ny, nz)
                        : new ArrayFloat.D3(nx, ny, nz);
                Index ima = array.getIndex();
                for (int i = 0; i < nx; i++) {
                    for (int j = 0; j < ny; j++) {
                        for (int k = 0; k < nz; k++) {
                            try {
                                ima.set(i, j, k);
                                Voxel voxel = voxelisation.getVoxel(i, j, k);
                                float value = (float) voxel.getFieldValue(Voxel.class, variable.getShortName(), voxel);
                                array.setFloat(ima, value);
                            } catch (SecurityException | NoSuchFieldException | IllegalAccessException ex) {
                                LOGGER.error(null, ex);
                            }
                        }
                    }
                }
                try {
                    writer.write(ncvar, array);
                } catch (InvalidRangeException ex) {
                    LOGGER.error(null, ex);
                }
            }
        }

        writer.close();

        return new File[]{outputFile};
    }
}
